/*
 * V6 disk editor
 *
 * Russ Cox <rsc@mit.edu> September 2003 [sic]
 */

#define _BSD_EXTENSION
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>

#ifndef PLAN9
#define USED
#endif

/* It's too hard to know who has typedefed what. */
#define uint16 _v6_uint16
#define uint _v6_uint
#define uchar _v6_uchar


typedef unsigned short uint16;
typedef unsigned int uint;
typedef unsigned char uchar;

#define nelem(x) (sizeof(x)/sizeof((x)[0]))

enum {
	BSIZE	= 512,
	SUPERB	= 1,
	ISIZE	= 32,
	INO2B	= BSIZE/ISIZE,
	ROOTINUM= 1,
	NAMELEN	= 14,
	DSIZE	= 2+NAMELEN,
	DIR2B	= BSIZE/DSIZE,
	NUM2B	= BSIZE/2,
	NADDR	= 8,
	NFREE	= 100,
};

typedef struct Super Super;
struct Super {
	uint16 isize;
	uint16 fsize;
	uint16 nfree;
	uint16 free[NFREE];
	uint16 ninode;
	uint16 inode[NFREE];
	uchar flock;
	uchar ilock;
	uchar fmod;
	uchar ronly;
	uint16 time[2];
};

typedef struct Inode Inode;
struct Inode {
	uint16 mode;
	uchar nlink;
	uchar uid;
	uchar gid;
	uint size;
	uint16 addr[NADDR];
	uint16 atime[2];
	uint16 mtime[2];
	uint16 inum;	/* not on disk */
};

enum {	/* Inode.mode */
	IALLOC	= 0100000,
	IFMT	= 0060000,
	IFDIR	= 0040000,
	IFCHR	= 0020000,
	IFBLK	= 0060000,
	ILARG	= 0010000,
	ISUID	= 0004000,
	ISGID	= 0002000,
	ISVTX	= 0001000,
	IREAD	= 0000400,
	IWRITE	= 0000200,
	IEXEC	= 0000100,
};

typedef struct Dir Dir;
struct Dir {
	uint16 inum;
	char name[NAMELEN+1];
};

typedef struct Free Free;
struct Free {
	uint16 blk[NFREE];
};

uchar *disk;
uint ndisk;
Super super;
char err[256];
uint curinum = 1;	/* the root */
int diskfd;

void*
getblk(uint bno)
{
	if(bno >= ndisk/BSIZE){
		fprintf(stderr, "attempt to access past end of disk bno=%d\n", bno);
		*(int*)0=0;
		exit(1);
	}
	return disk+bno*BSIZE;
}

uint16
getint(uchar *p, uint off)
{
	return p[off*2] | (p[off*2+1] << 8);
}

void
putint(uchar *p, uint off, uint val)
{
	p[off*2] = val;
	p[off*2+1] = val>>8;
}

long
readn(int f, void *av, long n)
{
	char *a;
	long m, t;

	a = av;
	t = 0;
	while(t < n){
		m = read(f, a+t, n-t);
		if(m <= 0){
			if(t == 0)
				return m;
			break;
		}
		t += m;
	}
	return t;
}

void
readdisk(char *name)
{
	struct stat s;

	if((diskfd = open(name, O_RDWR)) < 0){
		fprintf(stderr, "open %s", name);
		perror("");
		fprintf(stderr, "\n");
		exit(1);
	}
	
	if(fstat(diskfd, &s) < 0){
		fprintf(stderr, "cannot stat %s", name);
		perror("");
		fprintf(stderr, "\n");
		exit(1);
	}

	if(s.st_size < 1024 || s.st_size > 4*1024*1024){
		fprintf(stderr, "bad disk size %d\n", s.st_size);
		exit(1);
	}

	ndisk = s.st_size;
	disk = malloc(ndisk);
	if(disk == 0){
		perror("malloc");
		fprintf(stderr, "\n");
		exit(1);
	}

	if(readn(diskfd, disk, ndisk) != ndisk){
		perror("read");
		fprintf(stderr, "\n");
		exit(1);
	}
}

void
writedisk(void)
{
	if(lseek(diskfd, 0, 0) != 0
	|| write(diskfd, disk, ndisk) != ndisk){
		perror("writedisk");
		fprintf(stderr, "\n");
		exit(1);
	}
}

void
readsuper(void)
{
	int i;
	uchar *b;

	b = getblk(SUPERB);
	super.isize = getint(b, 0);
	super.fsize = getint(b, 1);
	super.nfree = getint(b, 2);
	for(i=0; i<100; i++)
		super.free[i] = getint(b, 3+i);
	super.ninode = getint(b, 103);
	for(i=0; i<100; i++)
		super.inode[i] = getint(b, 104+i);
	super.flock = b[204*2];
	super.ilock = b[204*2+1];
	super.fmod = b[205*2+0];
	super.ronly = b[205*2+1];
	super.time[0] = getint(b, 206);
	super.time[1] = getint(b, 207);
}

void
writesuper(void)
{
	int i;
	uchar *b;

	b = getblk(SUPERB);
	putint(b, 0, super.isize);
	putint(b, 1, super.fsize);
	putint(b, 2, super.nfree);
	for(i=0; i<100; i++)
		putint(b, 3+i, super.free[i]);
	putint(b, 103, super.ninode);
	for(i=0; i<100; i++)
		putint(b, 104+i, super.inode[i]);
	b[204*2] = super.flock;
	b[204*2+1] = super.ilock;
	b[205*2] = super.fmod;
	b[205*2+1] = super.ronly;
	putint(b, 206, super.time[0]);
	putint(b, 207, super.time[1]);
}

void
initfs(void)
{
	readsuper();

	if(super.fsize > ndisk/BSIZE)
		fprintf(stderr, "warning: disk superblock claims more blocks than file\n");

	if(1+1+super.isize > ndisk/BSIZE
	|| super.isize >= super.fsize){
		fprintf(stderr, "error: disk superblock claims more inode blocks than there are disk blocks (!)\n");
		exit(1);
	}

	if(super.fsize < 3 || super.isize < 1){
		fprintf(stderr, "error: disk too small\n");
		exit(1);
	}
}

Inode
iget(uint inum)
{
	int i;
	uchar *b;
	Inode ino;

	memset(&ino, 0, sizeof ino);
	ino.inum = inum;
	inum--;
	b = getblk(SUPERB+1+inum/INO2B);
	b += (inum%INO2B)*ISIZE;
	ino.mode = getint(b, 0);
	ino.nlink = b[2];
	ino.uid = b[3];
	ino.gid = b[4];
	ino.size = (b[5]<<16)|getint(b, 3);
	for(i=0; i<8; i++)
		ino.addr[i] = getint(b, 4+i);
	ino.atime[0] = getint(b, 12);
	ino.atime[1] = getint(b, 13);
	ino.mtime[0] = getint(b, 14);
	ino.mtime[1] = getint(b, 15);
	return ino;
}

void
iput(Inode *ino)
{
	int i;
	uchar *b;
	uint inum;

	inum = ino->inum-1;
	b = getblk(SUPERB+1+inum/INO2B);
	b += (inum%INO2B)*ISIZE;
	putint(b, 0, ino->mode);
	b[2] = ino->nlink;
	b[3] = ino->uid;
	b[4] = ino->gid;
	b[5] = ino->size>>16;
	putint(b, 3, ino->size);
	for(i=0; i<8; i++)
		putint(b, 4+i, ino->addr[i]);
	putint(b, 12, ino->atime[0]);
	putint(b, 13, ino->atime[1]);
	putint(b, 14, ino->mtime[0]);
	putint(b, 15, ino->mtime[1]);
}

Dir
getdir(uchar *p, uint off)
{
	Dir d;

	p += off*DSIZE;
	d.inum = getint(p, 0);
	memmove(d.name, p+2, NAMELEN);
	d.name[NAMELEN] = 0;
	return d;
}

void
putdir(uchar *p, uint off, Dir d)
{
	p += off*DSIZE;
	putint(p, 0, d.inum);
	memmove(p+2, d.name, NAMELEN);
}

uint16
alloc(void)
{
	int i;
	uchar *bp;
	uint bno;

	if(super.nfree == 0){
		strcpy(err, "out of blocks");
		return 0;
	}

	bno = super.free[--super.nfree];
	bp = getblk(bno);
	if(super.nfree == 0){
		for(i=0; i<NFREE; i++)
			super.free[i] = getint(bp, i);
		super.nfree = NFREE;
	}
	memset(bp, 0, BSIZE);
	return bno;
}

void
bfree(uint bno)
{
	int i;
	uchar *bp;

	if(super.nfree == NFREE){
		bp = getblk(bno);
		for(i=0; i<NFREE; i++)
			putint(bp, i, super.free[i]);
		super.nfree = 0;
	}
	super.free[super.nfree++] = bno;
}

uint16
ialloc(void)
{
	int i, j;
	uchar *bp;
	Inode ino;

	if(super.ninode == 0){
		/* scan for free inodes on disk */
		for(i=0; i<super.isize; i++){
			bp = getblk(SUPERB+1+i);
			for(j=0; j<INO2B; j++){
				if(bp[j*ISIZE+1]&(IALLOC>>8))
					continue;
				super.inode[super.ninode++] = 1+i*INO2B+j;
				if(super.ninode == NFREE)
					goto done;
			}
		}
		if(super.ninode == 0){
			strcpy(err, "out of inodes");
			return 0;
		}
	}
done:
	i = super.inode[--super.ninode];
	memset(&ino, 0, sizeof ino);
	ino.inum = i;
	ino.mode |= IALLOC;
	iput(&ino);
	return i;
}

void
itrunc(Inode *ino)
{
	int i, j, k;
	uchar *bp;

	for(i=0; i<NADDR; i++){
		if(ino->addr[i] == 0)
			continue;
		if(ino->mode&ILARG){
			bp = getblk(ino->addr[i]);
			for(j=0; j<NUM2B; j++){
				k = getint(bp, j);
				if(k != 0)
					bfree(k);
			}
		}
		bfree(ino->addr[i]);
		ino->addr[i] = 0;
	}
	ino->mode &= ~ILARG;
	ino->size = 0;
}

void
ifree(uint inum)
{
	Inode ino;

	ino = iget(inum);
	itrunc(&ino);
	ino.mode &= ~IALLOC;
	iput(&ino);

	if(super.ninode < NFREE)
		super.inode[super.ninode++] = inum;
}

uint16
bmap(Inode *ino, uint bno)
{
	int i;
	uchar *bp;
	uint nb;

	if((ino->mode&ILARG) == 0){
		if(bno < NADDR){
			nb = ino->addr[bno];
			if(nb == 0 && (nb = alloc()) != 0){
				ino->addr[bno] = nb;
				iput(ino);
			}
			return nb;
		}

		/* need to switch to large */
		if((nb = alloc()) == 0)
			return 0;

		bp = getblk(nb);
		for(i=0; i<NADDR; i++){
			putint(bp, i, ino->addr[i]);
			ino->addr[i] = 0;
		}
		ino->addr[0] = nb;
		ino->mode |= ILARG;
		iput(ino);

		/* fall through to large */
	}

	if(bno >= NUM2B*(NADDR-1)){
		/* punt on doubly indirect for now */
		strcpy(err, "file access too big");
		return 0;
	}

	/* follow first pointer */
	nb = ino->addr[bno/NUM2B];
	if(nb == 0){
		if((nb = alloc()) == 0)
			return 0;
		ino->addr[bno/NUM2B] = nb;
		iput(ino);
	}
	bp = getblk(nb);

	/* follow second pointer */
	nb = getint(bp, bno%NUM2B);
	if(nb == 0){
		if((nb = alloc()) == NULL)
			return 0;
		putint(bp, bno%NUM2B, nb);
	}
	return nb;
}

uint
_iwalk(Inode *ino, char *name, int dounlink)
{
	int i, j, b, n;
	Dir d;
	uchar *bp;

	n = (ino->size+BSIZE-1)/BSIZE;
	for(i=0; i<n; i++){
		b = bmap(ino, i);
		if(b == 0)
			continue;
		bp = getblk(b);
		for(j=0; j<DIR2B; j++){
			d = getdir(bp, j);
			if(d.inum && strcmp(name, d.name) == 0){
				n = d.inum;
				if(dounlink){
					d.inum = 0;
					putdir(bp, j, d);
				}
				return n;
			}
		}
	}
	return 0;
}

uint
iwalk(Inode *ino, char *name)
{
	return _iwalk(ino, name, 0);
}

uint
iunlink(Inode *ino, char *name)
{
	uint inum;
	Inode i2;

	inum = _iwalk(ino, name, 1);
	if(inum == 0)
		return 0;
	i2 = iget(inum);
	i2.nlink--;
	if(i2.nlink == 0)
		ifree(inum);
	return inum;
}

uint
ilink(Inode *dir, Inode *newi, char *name)
{
	int i, j, b, n;
	Dir d;
	uchar *bp;

	n = (dir->size+BSIZE-1)/BSIZE;
	for(i=0; i<n; i++){
		b = bmap(dir, i);
		if(b == 0)
			continue;
		bp = getblk(b);
		for(j=0; j<DIR2B; j++){
			d = getdir(bp, j);
			if(d.inum == 0){
				newi->nlink++;
				iput(newi);
				d.inum = newi->inum;
				strncpy(d.name, name, NAMELEN);
				putdir(bp, j, d);
				if(i==n-1 && i*BSIZE+(j+1)*DSIZE > dir->size){
					dir->size = i*BSIZE+(j+1)*DSIZE;
					iput(dir);
				}
				return d.inum;
			}
		}
	}

	b = bmap(dir, n);
	if(b == 0)
		return 0;
	dir->size = n*BSIZE+DSIZE;
	iput(dir);

	newi->nlink++;
	iput(newi);
	bp = getblk(b);
	d.inum = newi->inum;
	strncpy(d.name, name, NAMELEN);
	putdir(bp, 0, d);
	return d.inum;
}

uint
namei(char *path, Inode *parent, char **lastelem)
{
	static char elem[NAMELEN+1];
	char *q;
	uint inum, lastinum;
	Inode ino;

	if(lastelem)
		*lastelem = 0;
	if(parent)
		memset(parent, 0, sizeof *parent);
	inum = ROOTINUM;
	lastinum = inum;
	strcpy(elem, "/");	/* for lastelem */
	while(*path){
		while(*path=='/')
			path++;
		if(*path == 0)
			break;
		for(q=path+1; *q!='/' && *q; q++)
			;
		if(q-path > NAMELEN){
			strcpy(err, "path element too long");
			return 0;
		}
		memmove(elem, path, q-path);
		elem[q-path] = 0;
		path = q;

		lastinum = inum;
		ino = iget(inum);
		inum = iwalk(&ino, elem);
		if(inum == 0)
			break;
	}
	if(*path == 0){
		if(lastelem)
			*lastelem = elem;
		if(parent)
			*parent = iget(lastinum);
	}
	return inum;
}

int
iread(Inode *ino, uchar *data, uint n, uint off)
{
	uchar *bp;
	uint b, o, fo, fn;

	if(off >= ino->size)
		return 0;
	if(off+n > ino->size)
		n = ino->size - off;

	o = n;
	while(n > 0){
		b = bmap(ino, off/BSIZE);
		if(b == 0)
			break;
		bp = getblk(b);
		fo = off%BSIZE;
		fn = BSIZE - fo;
		if(fn > n)
			fn = n;
		memmove(data, bp+fo, fn);
		data += fn;
		n -= fn;
		off += fn;
	}
	return o - n;
}

int
iwrite(Inode *ino, uchar *data, uint n, uint off)
{
	uchar *bp;
	uint b, o, fo, fn;

	o = n;
	while(n > 0){
		b = bmap(ino, off/BSIZE);
		if(b == 0)
			break;
		bp = getblk(b);
		fo = off%BSIZE;
		fn = BSIZE - fo;
		if(fn > n)
			fn = n;
		memmove(bp+fo, data, fn);
		data += fn;
		n -= fn;
		off += fn;
	}
	if(off > ino->size){
		ino->size = off;
		iput(ino);
	}
	return o - n;
}

void
cp1(char *src, char *dst)
{
	uint inum;
	Inode dstino, parentino;
	Inode srcino;
	int dstfd, srcfd;
	int dstisdir;
	struct stat st;
	char buf[500], *elem, *p;
	uchar buf1[4096];
	int n, tot, m;

	srcfd = dstfd = -1;
	dstisdir = 0;
	if(dst[0] != ':'){
		inum = namei(dst, 0, 0);
		if(inum){
			dstino = iget(inum);
			if((dstino.mode&IFMT) == IFDIR)
					dstisdir = 1;
		}
	}else{
		if(stat(dst+1, &st) >= 0 && S_ISDIR(st.st_mode))
			dstisdir = 1;
	}

	if(src[0] != ':'){
		inum = namei(src, 0, 0);
		if(inum == 0){
			printf("cp -- %s %s\n", src, err);
			return;
		}
		srcino = iget(inum);
		if(srcino.mode&IFMT){
			printf("cp -- %s is special\n", src);
			return;
		}
	}else{
		if(stat(src+1, &st) < 0){
			printf("cp -- stat %s: %s\n", src, strerror(errno));
			return;
		}
		if((st.st_mode&S_IFMT) != S_IFREG){
			printf("cp -- %s is special\n", src);
			return;
		}
		if((srcfd = open(src+1, O_RDONLY)) < 0){
			printf("cp -- cannot open %s: %s\n", src, strerror(errno));
			return;
		}
	}

	if(dstisdir){
		p = src+strlen(src)-1;
		while(p>src && *p=='/')
			p--;
		while(p>src && *p!='/')
			p--;
		if(dst[strlen(dst)-1] == '/')
			p++;
		sprintf(buf, "%s%s", dst, p);
		dst = buf;
	}

	if(dst[0] != ':'){
		if((inum = namei(dst, &parentino, &elem)) == 0){
			if(!elem){
				printf("cp -- creating %s: %s\n", dst, err);
				goto out;
			}
			inum = ialloc();
			if(inum == 0){
				printf("cp -- %s\n", err);
				goto out;
			}
			dstino = iget(inum);
			if(ilink(&parentino, &dstino, elem) == 0){
				printf("cp -- linking %s: %s\n", dst, err);
				ifree(inum);
				goto out;
			}
		}else{
			dstino = iget(inum);
		}
	}else{
		if((dstfd = creat(dst+1, 0666)) < 0){
			printf("cp -- cannot creat %s: %s\n", dst, strerror(errno));
			goto out;
		}
	}

	tot = 0;
	for(;;){
		if(srcfd >= 0)
			n = read(srcfd, buf1, sizeof buf1);
		else
			n = iread(&srcino, buf1, sizeof buf1, tot);
		if(n < 0){
			printf("cp -- reading %s: %s\n",
				src, srcfd >= 0 ? strerror(errno) : err);
			goto out;
		}
		if(n == 0)
			break;
		if(dstfd >= 0)
			m = write(dstfd, buf1, n);
		else
			m = iwrite(&dstino, buf1, n, tot);
		if(m != n){
			printf("cp -- writing %s: %s\n",
				dst, dstfd >= 0 ? strerror(errno) : err);
			goto out;
		}
		tot += n;
	}

out:
	if(dstfd >= 0)
		close(dstfd);
	if(srcfd >= 0)
		close(srcfd);
}

int
xcp(int argc, char **argv)
{
	int i;
	char *dst;
	Inode ino;
	uint inum;
	struct stat st;

	if(argc < 3)
		return -1;

	if(argc > 3){
		dst = argv[argc-1];
		if(dst[0] != ':'){
			inum = namei(dst, 0, 0);
			if(inum == 0){
				printf("cp -- %s does not exist\n", dst);
				return 0;
			}
			ino = iget(inum);
			if((ino.mode&IFMT) != IFDIR){
				printf("cp -- %s is not a directory\n", dst);
				return 0;
			}
		}else{
			if(stat(dst+1, &st) < 0){
				printf("cp -- %s does not exist\n", dst);
				return 0;
			}
			if(!S_ISDIR(st.st_mode)){
				printf("cp -- %s is not a directory\n", dst);
				return 0;
			}
		}
		for(i=1; i<argc-1; i++)
			cp1(argv[i], argv[argc-1]);
	}else{
		cp1(argv[1], argv[2]);
	}
	return 0;
}

int
xrm(int argc, char **argv)
{
	int i;
	Inode iparent;
	char *elem;
	uint inum;

	if(argc == 1)
		return -1;
	for(i=1; i<argc; i++){
		elem = 0;
		inum = namei(argv[i], &iparent, &elem);
		if(inum == 0){
			printf("rm %s -- %s\n", argv[i], err);
			continue;
		}
		iunlink(&iparent, elem);
	}
	return 0;
}

int
xchmod(int argc, char **argv)
{
	int i;
	uint inum, mode;
	Inode ino;

	if(argc < 3)
		return -1;

	mode = strtol(argv[1], 0, 8);
	for(i=2; i<argc; i++){
		if((inum = namei(argv[i], 0, 0)) == 0){
			printf("chmod %s -- %s\n", argv[i], err);
			continue;
		}
		ino = iget(inum);
		ino.mode = (ino.mode & ~07777) | (mode&07777);
		iput(&ino);
	}
	return 0;
}

char*
rwx(int m)
{
	static char *r[] = {
		"---", "--x", "-w-", "-wx",
		"r--", "r-x", "rw-", "rwx",
	};
	return r[m&7];
}

void
lsi(uint inum, int dflag, char *name)
{
	Inode ino;
	char buf[256];

	ino = iget(inum);
	if((ino.mode&IFMT)==IFDIR && !dflag){
		int i, j, b, n;
		Dir d;
		uchar *bp;
	
		n = (ino.size+BSIZE-1)/BSIZE;
		for(i=0; i<n; i++){
			b = bmap(&ino, i);
			if(b == 0)
				continue;
			bp = getblk(b);
			for(j=0; j<DIR2B; j++){
				d = getdir(bp, j);
				if(d.inum == 0)
					continue;
				sprintf(buf, "%s%s%s", name,
					name[strlen(name)-1]=='/' ? "" : "/",
					d.name);
				lsi(d.inum, 1, buf);
			}
		}
		return;
	}

	printf("%c%c%c%c%c%c%s%s%s nl=%-4d uid=%-4d gid=%-4d size=%-9d inum=%-6d %s\n",
		(ino.mode&IALLOC) ? 'a' : '-',
		(ino.mode&IFMT) == IFDIR ? 'd' :
			(ino.mode&IFMT) == IFCHR ? 'c' :
			(ino.mode&IFMT) == IFBLK ? 'b' : '-',
		(ino.mode&ILARG) ? 'l' : '-',
		(ino.mode&ISUID) ? 's' : '-',
		(ino.mode&ISGID) ? 'g' : '-',
		(ino.mode&ISVTX) ? 'v' : '-',
		rwx(ino.mode>>6),
		rwx(ino.mode>>3),
		rwx(ino.mode),
		ino.nlink,
		ino.uid,
		ino.gid,
		ino.size,
		inum,
		name);
}	

int
xls(int argc, char **argv)
{
	int i, dflag;
	uint inum;

	dflag = 0;
	if(argc > 1 && strcmp(argv[1], "-d") == 0){
		dflag = 1;
		argc--;
		argv++;
	}

	if(argc == 1)
		return -1;

	for(i=1; i<argc; i++){
		if((inum = namei(argv[i], 0, 0)) == 0){
			printf("ls %s -- %s\n", argv[i], err);
			continue;
		}
		lsi(inum, dflag, argv[i]);
	}
	return 0;
}

int
xmkdir(int argc, char **argv)
{
	char *elem;
	Inode iparent;
	uint inum;
	Inode ino;

	if(argc != 2)
		return -1;

	inum = namei(argv[1], &iparent, &elem);
	if(inum != 0){
		printf("mkdir -- %s already exists\n", argv[1]);
		return 0;
	}

	inum = ialloc();
	if(inum == 0){
		printf("mkdir -- %s\n", err);
		return 0;
	}

	ino = iget(inum);
	ino.mode = IALLOC|IFDIR|0777;

	if(ilink(&iparent, &ino, elem) == 0){
		printf("mkdir -- %s\n", err);
		ifree(inum);
	}

	return 0;
}

int
xmknod(int argc, char **argv)
{
	char *elem;
	Inode parent;
	int c, maj, min;
	uint inum;
	Inode ino;

	if(argc != 5)
		return -1;

	c = argv[1][0];
	if(c != 'c' && c != 'b')
		return -1;
	if(argv[1][1] != 0)
		return -1;

	maj = atoi(argv[2]);
	min = atoi(argv[3]);

	inum = namei(argv[4], &parent, &elem);
	if(inum != 0){
		printf("mknod -- %s already exists\n", argv[4]);
		return 0;
	}
	if(elem == 0){
		printf("mknod -- dir for %s does not exist\n", argv[4]);
		return 0;
	}

	inum = ialloc();
	if(inum == 0){
		printf("mknod -- %s\n", err);
		return 0;
	}
	ino = iget(inum);
	ino.mode = IALLOC|(c=='b'?IFBLK:IFCHR)|0666;
	ino.addr[0] = (maj<<8)|min;

	if(ilink(&parent, &ino, elem) == 0){
		printf("mknod -- %s\n", err);
		ifree(inum);
	}

	return 0;
}
	
typedef struct Cmd Cmd;
struct Cmd {
	char *name;
	int (*fn)(int, char**);
	char *usage;
};

int xhelp(int, char**);

Cmd cmdtab[] = {
	"chmod",xchmod,	"mode file... - change file modes",
	"cp",	xcp,	"[!]src... [!]dst -- copy files, ! is in real os",
	"help",	xhelp,	"-- list command summary",
	"ls",	xls,	"[-d] file... -- list files",
	"mkdir",xmkdir,	"path - create directory",
	"mknod",xmknod,	"c|b major minor path - create device file",
	"rm",	xrm,	"file... - remove files",
};

int
xhelp(int argc, char **argv)
{
	int i;

	USED(argc);
	USED(argv);
	for(i=0; i<nelem(cmdtab); i++)
		printf("%s %s\n", cmdtab[i].name, cmdtab[i].usage);
	return 0;
}

void
runcmd(char *s)
{
	int i, nf;
	char *f[10], *p;

	for(i=0; i<10; i++){
		if((p = strtok(s, " \t")) == NULL)
			break;
		s = NULL;
		f[i] = p;
	}
	nf = i;

	if(nf == 0)
		return;

	for(i=0; i<nelem(cmdtab); i++){
		if(strcmp(f[0], cmdtab[i].name) == 0){
			if((*cmdtab[i].fn)(nf, f) < 0)
				fprintf(stderr, "usage: %s %s\n",
					cmdtab[i].name, cmdtab[i].usage);
			return;
		}
	}
	fprintf(stderr, "?unknown cmd '%s' -- try help\n", f[0]);
}

void
usage(void)
{
	fprintf(stderr, "usage: de v6disk [cmd...]\n");
	exit(1);
}

int
main(int argc, char **argv)
{
	int i;
	int quiet;
	char buf[512];

	quiet = 0;
	if(argc > 1 && strcmp(argv[1], "-q") == 0){
		quiet = 1;
		argc--;
		argv++;
	}

	if(argc < 2)
		usage();

	readdisk(argv[1]);
	initfs();

	if(argc > 2){
		for(i=2; i<argc; i++){
			strcpy(buf, argv[i]);
			runcmd(buf);
		}
	}else{
		for(;;){
			if(!quiet) fprintf(stderr, ">>> ");
			fflush(stdout);
			fflush(stderr);
			if(fgets(buf, sizeof buf, stdin) == NULL)
				break;
			buf[strlen(buf)-1] = '\0';
			runcmd(buf);
		}
	}

	writesuper();
	writedisk();

	return 0;
}


